#include <Arduino.h>
#include "application.h"
#include "synth_wave.h"

const struct
{ uint16_t attack;  // msec
  uint16_t decay;   // msec
  float    sustain; // level
  uint16_t release; // msec
  uint16_t quelch;  // msec
} waves[ NBR_WAVE_EFFECTS ] =
{//  A    D    S    R    Q
  {  25,  50, 0.3,  10, 10 },
  {   5,  25, 0.3, 100, 10 },
  {  10, 100, 0.3, 250, 10 },
  { 100,  50, 0.5, 100, 10 },
  { 500,  10, 0.8, 250, 10 }
};

uint16_t synth_instrument_wave::milliseconds2count(uint32_t milliseconds) 
{
  milliseconds = ((milliseconds*AUDIO_SAMPLE_RATE/1000)+7) >> 3;
  if (milliseconds > 65535) milliseconds = 65535; // 11.88 seconds max
  return milliseconds;
}

//- - - - - - - - - - - - - -

void synth_instrument_wave::attack(uint16_t milliseconds) 
{
  attack_count = milliseconds2count(milliseconds);
  if (attack_count == 0) attack_count = 1;
}

void synth_instrument_wave::decay(uint16_t milliseconds) 
{
  decay_count = milliseconds2count(milliseconds);
  if (decay_count == 0) decay_count = 1;
}

void synth_instrument_wave::sustain(float level) 
{
  if (level < 0.0) level = 0;
  else if (level > 1.0) level = 1.0;
  sustain_mult = level * 1073741824.0;
}

void synth_instrument_wave::release(uint16_t milliseconds) 
{
  release_count = milliseconds2count(milliseconds);
  if (release_count == 0) release_count = 1;
}

void synth_instrument_wave::quelch(uint16_t milliseconds) 
{
  quelch_count = milliseconds2count(milliseconds);
  if (quelch_count == 0) quelch_count = 1;
}

//- - - - - - - - - - - - - -
          
void synth_instrument_wave::startInstrument( ui_controls_struct *ui, int key )
{
  // end current note
  if (env_state != ENV_IDLE) // end current note
  {
    noInterrupts();
      env_state = ENV_QUELCH;
      env_count = quelch_count;
      env_inc32 = (-env_mult32) / env_count;
    interrupts();
  }
  else
  {
    noInterrupts();
      env_state = ENV_QUELCH;
      env_count = 0;
      env_inc32 = 0;
    interrupts();
  }
  // pass startup parameters to QUELCH end state
  start_ui  = ui;
  start_key = key;
}

//- - - - - - - - - - - - - -

void synth_instrument_wave::stopInstrument()
{
  noInterrupts();
    env_state = ENV_IDLE;
  interrupts();
}

//- - - - - - - - - - - - - -

void synth_instrument_wave::release()
{
  if (env_state != ENV_IDLE && env_state != ENV_RELEASE) // release allowed in attack, decay, sustain
  {
    // DMA interrupt mutex  
    noInterrupts();
      env_state = ENV_RELEASE; // enter release state
      env_count = release_count;
      env_inc32 = -env_mult32 / (int32_t)env_count;
    interrupts();
  }
}

//- - - - - - - - - - - - - -

void synth_instrument_wave::updateInstrument( ui_controls_struct *ui )
{
  // proximity effect
  if (!ui->proximity_enabled)
    dds_tune = dds_tune0;    // optical effect is off
  else  
    if (ui->proximity > 5)   // minimum optical threshold
      dds_tune = dds_tune0 + ((int32_t)ui->proximity << 17);
    else
      dds_tune = dds_tune0;    
}

//- - - - - - - - - - - - - -
// called from DMA interrupt
byte synth_instrument_wave::runInstrument( int32_t *buf )
{
  if (env_state == ENV_IDLE) return 0;

  for (int32_t *end = buf + SOUND_BLOCK_SIZE; buf < end; )
  {
    if (env_count == 0)
      switch (env_state)
      {
        case ENV_ATTACK:  // attack completed
          env_state = ENV_DECAY; // enter decay phase
          env_count = decay_count;
          env_inc32 = (sustain_mult - 0x40000000) / (int32_t)env_count;
          break;

        case ENV_DECAY:   // decay completed
          env_state  = ENV_SUSTAIN; // enter sustain phase
          env_count  = 0xFFFF;
          env_mult32 = sustain_mult;
          env_inc32  = 0;
          break;

        case ENV_SUSTAIN: // sustain timed out, repeat sustain
          env_count = 0xFFFF;
          break;

        case ENV_RELEASE: // release completed
          env_state = ENV_IDLE; // envlope done
          return 1;       // there may be stuff in the current buffer

        case ENV_QUELCH:  // quelch completed

          // calc frequency for note
          frequency = NOTES[ (NUM_SCALES-1) - instrument[start_ui->wave_knob].notes_scale ].frequency[ start_key ];
          for (int m=0; m <= start_ui->band_knob + 1; m++) frequency *= 2;
      
          // set tuning rate
          dds_tune0 = pow(2,32)*frequency/AUDIO_SAMPLE_RATE;
          dds_phase = 0;
          updateInstrument( start_ui ); // adjust dds_tune for proximity effect
          
          // set envelope parameters
          byte e = start_ui->effect_knob;
          attack ( waves[ e ].attack );
          decay  ( waves[ e ].decay );
          sustain( waves[ e ].sustain );
          release( waves[ e ].release );
          quelch ( waves[ e ].quelch );
          
          // start envelope
          noInterrupts();
            env_state  = ENV_ATTACK; // enter attack phase
            env_count  = attack_count;
            env_mult32 = 0;
            env_inc32  = 0x40000000 / env_count;
          interrupts();
      }
    
    int32_t env_mult = env_mult32 >> 14;
    int32_t env_inc  = env_inc32  >> 17;
    
    env_mult  += env_inc;
    dds_phase += dds_tune;
    *(buf++) += (wavetable[ 0 ][ dds_phase >> 24 ] * env_mult) >> 16;
    env_mult  += env_inc;
    dds_phase += dds_tune;
    *(buf++) += (wavetable[ 0 ][ dds_phase >> 24 ] * env_mult) >> 16;
    env_mult  += env_inc;
    dds_phase += dds_tune;
    *(buf++) += (wavetable[ 0 ][ dds_phase >> 24 ] * env_mult) >> 16;
    env_mult  += env_inc;
    dds_phase += dds_tune;
    *(buf++) += (wavetable[ 0 ][ dds_phase >> 24 ] * env_mult) >> 16;
    env_mult  += env_inc;
    dds_phase += dds_tune;
    *(buf++) += (wavetable[ 0 ][ dds_phase >> 24 ] * env_mult) >> 16;
    env_mult  += env_inc;
    dds_phase += dds_tune;
    *(buf++) += (wavetable[ 0 ][ dds_phase >> 24 ] * env_mult) >> 16;
    env_mult  += env_inc;
    dds_phase += dds_tune;
    *(buf++) += (wavetable[ 0 ][ dds_phase >> 24 ] * env_mult) >> 16;
    env_mult  += env_inc;
    dds_phase += dds_tune;
    *(buf++) += (wavetable[ 0 ][ dds_phase >> 24 ] * env_mult) >> 16;

    env_mult32 += env_inc32;
    env_count--;
  }
  
  return 1; // running
}

